from enum import Enum


class Providers(Enum):
    OPENAI = "OPENAI"
    ANTHROPHIC = "ANTHPROPIC"
    VERTEX_AI = "VERTEX_AI"
    GPT4ALL = "GPT4ALL"
    OLLAMA = "OLLAMA"
    AZURE_OPENAI = "AZURE_OPENAI"
